/******************************************************************
**  This file is distributed under the GNU General Public License.
**  (c) 2005 Jakson Alves de Aquino 
*******************************************************************/

/* recode.c for STATIST */

/* WARNING: Although the ignore list includes ",", this program is working 
 properly only when the delimiters are either "\t" or " "
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NODATA 'M'

/* TODO: When a line in the config_file is too big the user should have the 
   option of using a "\" instead of writing a very large line */

struct Column{
  char * label; /* column label */
  char * newlabel;
  int pos;      /* position of the column */
  int n;        /* number of values to be recoded */
  char ** oldv; /* old value */
  char ** rgv1; /* begin of range of values to be recoded */ 
  char ** rgv2; /* end of range of values to be recoded */
  char ** newv; /* new value */
};

struct Column ** col; /* list of recoding information */
char ** dblab;        /* list of column labels in the original database */
int nlab = 0;         /* number of labels found */
int max = 1000;       /* maximum number of columns that can be recoded before 
                         new reallocation of memory */
int maxr = 10;        /* maximum number of recodifications for each column 
                         before new reallocation of memory */
int rlen = 10001;     /* amount of memory allocated to read rows */


void init_col(){
  col = (struct Column**) malloc(sizeof(struct Column*) * max);
  int i;
  for(i = 0; i < max; i++){
    col[i] = (struct Column*) malloc(sizeof(struct Column));
    col[i]->label = NULL;
    col[i]->newlabel = NULL;
    col[i]->pos = -1;
    col[i]->n = 0;
    col[i]->oldv = NULL;
    col[i]->rgv1 = NULL;
    col[i]->rgv2 = NULL;
    col[i]->newv = NULL;
  }
}

void delete_col(){
  int i, j;
  for(i = 0; i < nlab; i++){
    if(col[i]->label != NULL)
      free(col[i]->label);
    if(col[i]->newlabel != NULL)
      free(col[i]->newlabel);
    for(j = 0; j < col[i]->n; j++){
      if(col[i]->oldv[j] != NULL)
        free(col[i]->oldv[j]);
      if(col[i]->rgv1[j] != NULL)
        free(col[i]->rgv1[j]);
      if(col[i]->rgv2[j] != NULL)
        free(col[i]->rgv2[j]);
      if(col[i]->newv[j] != NULL)
        free(col[i]->newv[j]);
    }
    if(col[i]->oldv != NULL)
      free(col[i]->oldv);
    if(col[i]->rgv1 != NULL)
      free(col[i]->rgv1);
    if(col[i]->rgv2 != NULL)
      free(col[i]->rgv2);
    if(col[i]->newv != NULL)
      free(col[i]->newv);
  }
  free(col);
}

void printList(){
  int i, j;
  printf("Recodifications:\n");
  for(i = 0; i < nlab; i++)
    if(col[i]->label != NULL){
      printf("  %s", col[i]->label);
      for(j = 0; j < col[i]->n; j++){
        if(col[i]->oldv[j] != NULL)
          printf(" (%s", col[i]->oldv[j]);
        if(col[i]->rgv1[j] != NULL)
          printf(" (%s", col[i]->rgv1[j]);
        if(col[i]->rgv2[j] != NULL)
          printf("-%s", col[i]->rgv2[j]);
        if(col[i]->newv[j] != NULL)
          printf("=%s)", col[i]->newv[j]);
      }
      if(col[i]->newlabel != NULL)
        printf(" %s", col[i]->newlabel);
      printf("\n");
    }
}

void checkTheList(){
  /* look for inconsistences in the list */
  int i, j;
  int nErrors = 0;
  for(i = 0; i < nlab; i++)
    if(col[i]->label != NULL){
      for(j = 0; j < col[i]->n; j++){
        if(col[i]->oldv[j] == NULL && col[i]->rgv1[j] == NULL){
          printf("Error: Neither old value nor old value range defined for"
              "variable \"%s\", recodification %i\n", col[i]->label, (j+1));
          nErrors++;
        }
        if(col[i]->newv[j] == NULL){
          printf("Error: The new value wasn't defined for variable \"%s\", "
              "recodification %i\n", col[i]->label, (j+1));
          nErrors++;
        }
        if((col[i]->rgv1[j] != NULL && col[i]->rgv2[j] == NULL) ||
            (col[i]->rgv1[j] == NULL && col[i]->rgv2[j] != NULL)){
          printf("Error: Old value range not completely defined for variable "
              "%s, recodification %i\n", col[i]->label, (j+1));
          nErrors++;
        }
        if(col[i]->oldv[j] != NULL && 
            !(col[i]->rgv1[j] == NULL && col[i]->rgv2[j] == NULL)){
          printf("Error in the recodification information for variable %s\n",
              col[i]->label);
          nErrors++;
        }
      }
    }
  if(nErrors > 0){
    printf("\nPlease, correct the errors before trying again.\n\n");
    exit(0);
  }
}

void printError(char * filename, int linenumber, char * line, int pos){
  printf("Error in file \"%s\", line %i:\n\n%s", filename, linenumber, line);
  int k;
  for(k = 0; k < pos; k++) printf(" ");
  printf("^\n\n");
  printList();
  delete_col();
  exit(0);
}

void extractLabels(char * s1){
  char *s, *token;
  char **temp;
  const char ignore[] = " ,;\n\t";
  const char var_id[] = "#%";/* these chars indicate: here are the labels */
  int i;

  dblab = (char**) malloc(sizeof(char**) * max);
  s = s1+strlen(var_id);  /* jump over var_id */
  printf("\nColumns:\n");
  while ((token = strtok(s, ignore))!= NULL) {
    s = NULL;
    dblab[nlab] = (char*) malloc(sizeof(char) * (strlen(token) + 1));
    strcpy(dblab[nlab], token);
    printf("  %i = %s\n", nlab+1, dblab[nlab]);
    nlab++;
    if(nlab >= max){ /* inflate dblab */
      max *= 2;
      temp = (char**) malloc(sizeof(char**) * max);
      for(i = 0; i < nlab; i++) temp[i] = dblab[i];
      free(dblab);
      dblab = temp;
    }
  }
  printf("\n");
  if (nlab==0) printf("No variables found in s1!\n");
}

void putLabels(char *b1, char *b2){
  char *s, *token;
  const char ignore[] = " ,;\n\t";
  const char var_id[] = "#%";/* these chars indicate: labels are here */

  b2[0] = 0;
  strcat(b2, "#% ");
  s = b1+strlen(var_id);  /* jump over var_id */
  int j = 0;
  while ((token = strtok(s, ignore))!= NULL) {
    s = NULL;
    strcat(b2, token);
    strcat(b2, "\t");
    if(col[j]->newlabel != NULL){
      strcat(b2, col[j]->newlabel);
      strcat(b2, "\t");
    }
    j++;
  }
  strcat(b2, "\n");
  if (nlab==0) printf("No variables found in b1!\n");
}

void recodeRow(FILE * f2, char *b1, char *b2, int lread, char * filename){
  double value, rgv1, rgv2, oldv;
  char ignore[]= " \n\t\0", *ptr, *token, *rec;
  int i, j;

  if (strlen(b1) > 1 && b1[0] != '#'){
    j = 0;
    ptr = b1;
    b2[0] = 0;

    while ((token = strtok(ptr, ignore))!= NULL) {
      ptr = NULL;
      if(col[j]->newlabel != NULL){ /* copy the old column before recoding */
        strcat(b2, token);
        strcat(b2, "\t");
      }
      if ((token[0] == NODATA && strlen(token)==1) || col[j]->label == NULL){
        strcat(b2, token);
      } else {
        if (!(sscanf(token, "%lf", &value)==1)){
          printf("Warning: can't convert \"%s\" into real number: "
              "\"%s\", line %i.", token, filename, lread);
          strcat(b2, token);
        } else {
          rec = NULL;
          for(i = 0; i < col[j]->n; i++){
            if(col[j]->oldv[i] != NULL){
              sscanf(col[j]->oldv[i], "%lf", &oldv);
              if(value == oldv) rec = col[j]->newv[i];
            }
            else{
              sscanf(col[j]->rgv1[i], "%lf", &rgv1);
              sscanf(col[j]->rgv2[i], "%lf", &rgv2);
              if(value >= rgv1 && value <= rgv2) rec = col[j]->newv[i];
            }
          }
          if(rec) strcat(b2, rec);
          else strcat(b2, token);
        }
      }
      j++;
      if(j < nlab) strcat(b2, "\t");
    } 
    fprintf(f2, "%s\n", b2);
  }
}

void recode_usage(char * name){
  printf("\nUsage:\n"
      "%s [-rlen number] config_file data_base dest_file\n\n"
      "The limits are %i variables, and %i "
      "recodifications for each variable\n\n", name, max, maxr);
  printf("If the largest row in the datafile is larger than 10000 "
      "characteres,\nyou have to indicate its size with the option -rlen.\n"
      "You don't need to know the exact value. Just choose a number "
      "big enough.\n\n");
  exit(0);
}

int inflateCol(struct Column * c, int cmaxr){
  int newmax = 2 * cmaxr;
  int i;
  char ** temp;
  temp = (char**) malloc(sizeof(char*) * newmax);
  for(i = 0; i < cmaxr; i++) temp[i] = c->oldv[i];
  free(c->oldv);
  c->oldv = temp;
  temp = (char**) malloc(sizeof(char*) * newmax);
  for(i = 0; i < cmaxr; i++) temp[i] = c->rgv1[i];
  free(c->rgv1);
  c->rgv1 = temp;
  temp = (char**) malloc(sizeof(char*) * newmax);
  for(i = 0; i < cmaxr; i++) temp[i] = c->rgv2[i];
  free(c->rgv2);
  c->rgv2 = temp;
  temp = (char**) malloc(sizeof(char*) * newmax);
  for(i = 0; i < cmaxr; i++) temp[i] = c->newv[i];
  free(c->newv);
  c->newv = temp;
  return(newmax);
}

int main(int argc, char * argv[]){
  if(argc < 4) recode_usage(argv[0]);

  if(strcmp(argv[1], "-rlen") == 0){
    if(argc < 6) recode_usage(argv[0]);
    rlen = atoi(argv[2]);
    if(rlen == 0){
      printf("Could not convert \"%s\" into an integer. rlen was "
          "set to 10000.\n", argv[2]);
      rlen = 10001;
    }
    argv[1] = argv[3];
    argv[2] = argv[4];
    argv[3] = argv[5];
  }

  FILE * f1;
  FILE * f2;
  f1 = fopen(argv[2], "r");
  char b1[rlen];
  char b2[rlen];
  int lread = 0; /* last line read from original database */

  /* read all _comment lines of database and extract labels */
  while(fgets(b1, rlen , f1)){
    lread++;
    if(b1[0] == '#'){
      if(b1[1] == '%'){
        extractLabels(b1);
      }
    } else break;
  }
  fclose(f1);
  
  /* read config_file */
  FILE * pf;
  char s[200];
  char b[200];
  int i, j, k, l, lastp, sl;
  int ln = 0;
  int n;
  int cmaxr; /* current maximum number of recodifications */
  pf = fopen(argv[1], "r");
   if (pf == NULL){
     printf("Error opening file %s.", argv[1]);
     return(0);
   }
   init_col();
   while (fgets(s, 100 , pf)){
     ln++; /* line number*/
     if(s[0] == '#' || strlen(s) < 6) continue;
     sl = strlen(s);
     for(i = 0; i < sl; i++) if(s[i] == ')') lastp = i;
     if(lastp < 6) continue;
     i = 0;
     while((s[i] >= '0' && s[i] <= '9') || (s[i] >= 'A' && s[i] <= 'Z') ||
         (s[i] >= 'a' && s[i] <= 'z') || s[i] == '_' || 
         (s[i] >= '' && s[i] <= '')){
       b[i] = s[i];
       i++;
     }
     b[i] = 0;
     /* find label position*/
     n = -1;
     for(j = 0; j < nlab; j++){
       if(strcmp(dblab[j], b) == 0)
         n = j;
     }
     if(n == -1){
       printf("The label \"%s\" wasn't found in the database!\n", b);
       continue;
     }
     col[n]->label = (char*) malloc(i + 1);
     strcpy(col[n]->label, b);
     col[n]->oldv = (char**) malloc(sizeof(char*) * maxr);
     col[n]->newv = (char**) malloc(sizeof(char*) * maxr);
     col[n]->rgv1 = (char**) malloc(sizeof(char*) * maxr);
     col[n]->rgv2 = (char**) malloc(sizeof(char*) * maxr);
     for(k = 0; k < maxr; k++){
       col[n]->oldv[k] = NULL;
       col[n]->newv[k] = NULL;
       col[n]->rgv1[k] = NULL;
       col[n]->rgv2[k] = NULL;
     }
     k = 0;
     cmaxr = maxr;
     do{
       if(k == cmaxr) cmaxr = inflateCol(col[n], cmaxr);
       while(s[i] != '(' && i < lastp) i++;
       while(!((s[i] >= '0' && s[i] <= '9') || s[i] == NODATA || s[i] == '-') 
           && i < lastp)
         i++;
       j = 0;
       /* get old value or range */
       while(((s[i] >= '0' && s[i] <= '9') || s[i] == NODATA || s[i] == '-')
           && i < lastp){
         b[j] = s[i];
         i++; j++;
       }
       b[j] = 0;
       l = strlen(b);
       j = 0;
       if(s[i] == '='){  /* it's a single value */
         col[n]->oldv[k] = (char*) malloc(l + 1);
         strcpy(col[n]->oldv[k], b);
       } else if(s[i] == '>'){ /* it's a range of values */
         col[n]->rgv1[k] = (char*) malloc(l + 1);
         strcpy(col[n]->rgv1[k], b);
         while(!((s[i] >= '0' && s[i] <= '9') || s[i] == NODATA || s[i] == '-')
             && i < lastp)
           i++;
          while(((s[i] >= '0' && s[i] <= '9') || s[i] == NODATA || s[i] == '-')
              && i < lastp){
           b[j] = s[i];
           i++; j++;
         }
         b[j] = 0;
         l = strlen(b);
         col[n]->rgv2[k] = (char*) malloc(l + 1);
         strcpy(col[n]->rgv2[k], b);
       } else continue;
       j = 0;
       while(!((s[i] >= '0' && s[i] <= '9') || s[i] == NODATA || s[i] == '-')
           && i < lastp)
         i++;
       while(((s[i] >= '0' && s[i] <= '9') || s[i] == NODATA || s[i] == '-')
           && i < lastp){
         b[j] = s[i];
         i++; j++;
       }
       b[j] = 0;
       l = strlen(b);
       col[n]->newv[k] = (char*) malloc(l + 1); /* get new value */
       strcpy(col[n]->newv[k], b);
       k++;
       col[n]->n = k;
     } while(i < lastp);
     /* recode into same variable or in different variable? */
     while((s[i] == ')' || s[i] == ' ') && i < sl) i++;
     if(i < (sl - 2)){ /* it seems that we have a new label */
       j = 0;
       while((s[i] >= '0' && s[i] <= '9') || (s[i] >= 'A' && s[i] <= 'Z') ||
         (s[i] >= 'a' && s[i] <= 'z') || s[i] == '_' || 
         (s[i] >= '' && s[i] <= '')){
         b[j] = s[i];
         i++; j++;
       }
       b[j] = 0;
       l = strlen(b);
       col[n]->newlabel = (char*) malloc(l + 1);
       strcpy(col[n]->newlabel, b);
     }
     n++;
   }
   fclose (pf);

  checkTheList();
  printList();

  
  /* TODO: instead of using the option -rlen, discover the database rows 
     length, and allocate memory to read them */

  /* put column labels in the new database */
  lread = 0;
  f1 = fopen(argv[2], "r");
  f2 = fopen(argv[3], "w");
  while(fgets(b1, rlen , f1)){
    lread++;
    if(b1[0] == '#'){
      if(b1[1] == '%'){
        putLabels(b1, b2);
        fputs(b2, f2);
      } else fputs(b1, f2);
    } else break;
  }
  printf("\nRecoding...\n");

  /* read database, recode row, and write to new database */
  recodeRow(f2, b1, b2, lread, argv[2]);
  while(fgets(b1, rlen , f1)){
    recodeRow(f2, b1, b2, lread, argv[2]);
    lread ++;
  }
    
  /* finish */
  fclose(f1);
  fclose(f2);
  delete_col();
  return(0);
}
