/******************************************************************
**  This file is distributed under the GNU General Public License.
**  (c) 2005 Jakson Alves de Aquino 
*******************************************************************/

/* xcols.c for STATIST */ 

/* "xcols" means "extract columns from fixed width data file" */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

char ** label;
int * begin;
int * end;
int n;
int max = 1000;
int rlen = 10001;     /* amount of memory to be allocated to read rows */

#define NODATA 'M'   /* Symbol that indicates missing values in data files */


void formatNumber(char * number, char * result){
  if(number[0] == ' '){  /* eliminating initial empty spaces */
    int j, l;
    do{
      l = strlen(number);
      for(j = 0; j < l; j++) number[j] = number[j+1];
    } while(number[0] == ' ');
  }
  if(strlen(number) == 0)  /* the value is missing */
    sprintf(result, "%c\t", NODATA);
  else{
    double n = atof(number);
    if(n == floor(n)){
      if(n > -2147483647 && n < 2147483647)
        sprintf(result, "%li\t", (long)n);
      else sprintf(result, "%s\t", number);
    }
    else sprintf(result, "%10.5e\t", n);
  }
}

void xcols_usage(char * name){
  printf("\nUsage:\n"
      "%s [-rlen number] config_file data_base dest_file\n\n", name);
  printf("If the largest row in the datafile is larger than 10000 "
      "characteres,\nyou have to indicate its size with the option -rlen.\n"
      "You don't need to know the exact value. Just choose a number "
      "big enough.\n\n");
  exit(0);
}

int main(int argc, char * argv[]){
  if(argc < 4) xcols_usage(argv[0]);
  
  if(strcmp(argv[1], "-rlen") == 0){
    if(argc < 6) xcols_usage(argv[0]);
    rlen = atoi(argv[2]) + 1;
    if(rlen == 0){
      printf("Could not convert \"%s\" into an integer. rlen was "
          "set to 10000.\n", argv[2]);
      rlen = 10001;
    }
    argv[1] = argv[3];
    argv[2] = argv[4];
    argv[3] = argv[5];
  }

  /* read config_file */
  FILE * pf;
  char s[1000];
  char b[1000];
  int i, j;
  pf = fopen(argv[1], "r");
   if (pf == NULL){
     printf("Error opening file %s.", argv[1]);
     return(0);
   }
   else {
     label = (char**) malloc(max);
     begin = (int*) malloc(max);
     end = (int*) malloc(max);
     while (fgets(s, 1000 , pf)){
       if(s[0] == '#' || strlen(s) < 3) 
         continue;
       i = 0;
       while((s[i] >= '0' && s[i] <= '9') || (s[i] >= 'A' && s[i] <= 'Z') ||
           (s[i] >= 'a' && s[i] <= 'z') || s[i] == '_' || 
           (s[i] >= '' && s[i] <= '')){
         b[i] = s[i];
         i++;
       }
       b[i] = 0;
       label[n] = (char*) malloc(i + 1);
       strcpy(label[n], b);
       while(!(s[i] >= '0' && s[i] <= '9')) i++;
       j = 0;
       while(s[i] >= '0' && s[i] <= '9'){
         b[j] = s[i];
         i++; j++;
       }
       b[j] = 0;
       begin[n] = atoi(b) - 1;
       while(!(s[i] >= '0' && s[i] <= '9')) i++;
       j = 0;
       while(s[i] >= '0' && s[i] <= '9'){
         b[j] = s[i];
         i++; j++;
       }
       b[j] = 0;
       end[n] = atoi(b) - 1;
       n++;
     }
     fclose (pf);
   }

  /* TODO: instead of using the option -rlen, discover the database rows 
     length, and allocate memory to read them */
  char b1[rlen];
  char b2[rlen];
  char number[50];
  char number2[50];
  unsigned int l;
  for(i = 0; i < rlen; i++) b2[i] = 0;

  /* read from origin, and write to destination */
  FILE * f1;
  FILE * f2;
  f1 = fopen(argv[2], "r");
  f2 = fopen(argv[3], "w");
  strcpy(b2, "#% ");  /* put labels */
  for(i = 0; i < n; i++){
    strcat(b2, label[i]);
    strcat(b2, "\t");
  }
  l = strlen(b2);
  b2[l - 1] = '\n'; /* replace the last '\t' with '\n' */
  fputs(b2, f2);
  for(i = 0; i < l; i++) b2[i] = 0;
  int k;
  while (fgets(b1, rlen, f1)){
    for(i = 0; i < n; i++){
      k = 0;
      for(j = begin[i]; j <= end[i]; j++){
        number[k] = b1[j];
        k++;
      }
      number[k] = 0;
      formatNumber(number, number2);
      strcat(b2, number2);
    }
    l = strlen(b2);
    b2[l - 1] = '\n';
    fputs(b2, f2);
    for(i = 0; i < l; i++) b2[i] = 0;
  }
  fclose(f1);
  fclose(f2);
    
  /* finish */
  return(0);
}
