/******************************************************************
**  This file is distributed under the GNU General Public License.
**  (c) 2005 Jakson Alves de Aquino 
*******************************************************************/

/* xsample.c for STATIST */ 

/* "xsample" means "extract a random sample of rows from data file" */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

int rlen = 10001;     /* amount of memory to be allocated to read rows */

void xsample_usage(char * name){
  printf("\nThis program extracts a random sample of rows from a given "
      "data file.\n\n");
  printf("\nUsage:\n\n"
      "  %s -s percentage [-rlen number] data_base dest_file\n\n"
      "where \"percentage\" is an integer between 1 and 99.\n\n", name);
  printf("If the largest row in the datafile is larger than 10000 "
      "characteres,\nyou have to indicate its size with the option -rlen.\n"
      "You don't need to know the exact value. Just choose a number "
      "big enough.\n\n");
  exit(0);
}

int main(int argc, char * argv[]){
  if(argc < 5) xsample_usage(argv[0]);

  int percent = -1;
  if(strcmp(argv[1], "-s") == 0){
    percent = atoi(argv[2]);
    if(percent > 99 || percent < 1){
      printf("\nError: \"%s\" is not a valid value for percentage\n\n", 
          argv[2]);
      exit(0);
    }
  } else xsample_usage(argv[0]);

  if(strcmp(argv[3], "-rlen") == 0){
    if(argc < 7) xsample_usage(argv[0]);
    rlen = atoi(argv[2]) + 1;
    if(rlen == 0){
      printf("Could not convert \"%s\" into an integer. rlen was "
          "set to 10000.\n", argv[4]);
      rlen = 10001;
    }
    argv[3] = argv[5];
    argv[4] = argv[6];
  }

  printf("Creating a new database with a random sample of approximately "
      "%i%% of \"%s\" rows...\n", percent, argv[3]);
  
  /* TODO: instead of using the option -rlen, discover the database rows 
     length, and allocate memory to read them */
  char s[rlen];

  /* read from source, and write to destine */
  srand(time(NULL));
  int k = percent * 10;
  int i;
  FILE * f1;
  FILE * f2;
  f1 = fopen(argv[3], "r");
  if(f1 == NULL){
    printf("Error: couldn't open file \"%s\"!", argv[3]);
    exit(0);
  }
  f2 = fopen(argv[4], "w");
  if(f2 == NULL){
    printf("Error: couldn't create file \"%s\"!", argv[4]);
    fclose(f1);
    exit(0);
  }
  while(fgets(s, rlen, f1)){
    if(s[0] == '#') fputs(s, f2);
    else{
      i = rand() % 1000;
      if(i < k)
        fputs(s, f2);
    }
  }
    
  /* finish */
  fclose(f1);
  fclose(f2);
  return(0);
}
